import java.awt.Point;


/** A Configuration represents an arrangement of pieces within the box. */

public class Configuration
{
    PieceList       pieces;         // the pieces in the configuration
    boolean         usesGoal;       // is the goal piece among them?
    
    int             distance;       // these 2 fields are used for BFS, with the configuration treated as
    Configuration   predecessor;    // a vertex in the graph of reachable configurations.  See CLRS, p. 532.

    int             movedPiece;     // this is additional predecessor information - how we got here from the 
    DirectionList   movedShifts;    // predecessor configuration.  Used by printSolution.

    public static int examined;
    
    /** the input specifications */
    
    public static InputSpec spec;
    
    
    /** maximum required solution length found so far */
    
    public static int maxDistance;
    
    
    /** Make a new, empty configuration. */
    
    public Configuration()
    {
        movedPiece = -1;
    }
    
    
    /** Make a copy of an existing configuration. */

    public Configuration(Configuration config)
    {
        pieces = config.pieces != null ? new PieceList(config.pieces) : null;
        usesGoal = config.usesGoal;
        distance = config.distance;
        predecessor = config.predecessor;
        movedPiece = config.movedPiece;
        movedShifts = config.movedShifts;
    }
    
                    
    /** Try all possible piece placements at (x, y) (including 'filling' it with
        empty space), and for each one, recursively try to fill the next spot in
        the box.  If we've thus filled the entire box, we have a candidate puzzle
        layout; see if it can be solved. */

    public void puzzleSearch(int x, int y)
    {
        Point where = new Point(x, y);
        
        if (y < spec.boxShape.height())
        {
            // Try all piece placements at (x, y).  For i == spec.pieceShapes.length, we choose to
            // make this mean fill square (x, y) with nothing.
            
            int nextx = (x + 1) % spec.boxShape.width();
            int nexty = y + (nextx == 0 ? 1 : 0);
            
            for (int i = 0; i <= spec.pieceShapes.length; ++i)
                if (i == spec.pieceShapes.length || placePiece(i, where))
                {
                    puzzleSearch(nextx, nexty);
                    
                    // Now take the last piece out, so we can try the next.
                    
                    if (i < spec.pieceShapes.length)
                        removePiece();
                }
        }
        else
            solve();                // We've filled the whole box.  Try to solve this puzzle.
    }
    
    
    /** Can this configuration be solved by making some sequence of moves?
        If so, and the minimum-length solution is longer than the longest found so
        far, print it out. */

    public void solve()
    {
        if (!usesGoal)              // Can't solve it if it doesn't even contain the goal piece!
            return;
        
        ++examined;
        
        Queue queue = new Queue();
        ConfigurationList seen = null;
        
        // Do a BFS from (a copy of) this configuration, looking for a solved state.
        // See CLRS, p. 532.  We don't use an explicit 'color' field; instead, we keep
        // a list 'seen' of all configurations we've already visited.
        
        queue.enqueue(new Configuration(this));
            
        while (!queue.empty())
        {
            Configuration config = queue.dequeue();

            seen = new ConfigurationList(config, seen);         // Add this config to the seen list.
            
            // Find all configurations reachable from config by moving a single piece.
            
            PieceList pl;
            int pieceNum;
            
            for (pl = config.pieces, pieceNum = 0; pl != null; pl = pl.next, ++pieceNum)
                for (ConfigurationList moves = config.findMoves(pieceNum); moves != null; moves = moves.next)
                {
                    Configuration newconfig = moves.config;
                    
                    // Have we already looked at this configuration?
                    
                    if ((seen == null || !seen.isPresent(newconfig)) && !queue.isPresent(newconfig))
                    {
                        // Found a new reachable configuration.  Check to see if it's solved, and add it to the queue.
                        
                        newconfig.distance = config.distance + 1;
                        newconfig.predecessor = config;
                        newconfig.movedPiece = pieceNum;
                        
                        if (newconfig.solved())
                        {                           
                            if (newconfig.distance > maxDistance)
                            {
                                // This is the most difficult puzzle we've found yet.
                                
                                maxDistance = newconfig.distance;

                                newconfig.printSolution();
                            }
                            
                            // We know the puzzle can be solved at this depth; no use looking any farther.
                            
                            return;
                        }
                        
                        // The new configuration is not solved; add it to the search queue.
                        
                        queue.enqueue(newconfig);
                    }                   
                }
        }
    }


    /** Return a list of all configurations reachable from this one by a sequence of shifts of a given piece.
        To do this, we do another BFS. */

    ConfigurationList findMoves(int pieceNum)
    {
        Queue queue = new Queue();
        ConfigurationList seen = null;
        
        Configuration start = new Configuration(this);
        start.movedShifts = null;
        
        queue.enqueue(start);

        while (!queue.empty())
        {
            Configuration config = queue.dequeue();
            seen = new ConfigurationList(config, seen);
        
            Piece piece = config.pieces.index(pieceNum);
            
            for (int dir = 0; dir < 4; ++dir)
                if (piece.canSlide(dir, config))
                {
                    // We found a piece that can slide.
                    
                    Configuration newconfig = new Configuration(config);
                    
                    newconfig.pieces.index(pieceNum).shift(dir);
                    
                    // Have we not seen this one yet?
                    
                    if ((seen == null || !seen.isPresent(newconfig)) && !queue.isPresent(newconfig))
                    {
                        // Found a new reachable configuration.
                        // Add it to the queue.

                        newconfig.movedShifts = new DirectionList(dir, newconfig.movedShifts);
                        queue.enqueue(newconfig);   
                    }                   
                }   
        }
        
        return seen;
    }

    /** Return a list of all configurations reachable by a sequence of shifts of a given piece. */
/*

    ConfigurationList tryShifts(int pieceNum, ConfigurationList done, DirectionList shifts)
    {
        done = new ConfigurationList(this, done);           // Add this configuration to the list.

        Piece piece = pieces.index(pieceNum);
        
        for (int dir = 0; dir < 4; ++dir)
            if (piece.canSlide(dir, this))
            {
                // We found a piece that can slide.
                
                Configuration newconfig = new Configuration(this);
                
                newconfig.pieces.index(pieceNum).shift(dir);
                
                // Is this configuration not already in our output list?
                
                if (done == null || !done.isPresent(newconfig))
                {
                    // Found a new reachable configuration.
                    // Try some more shifts from this new configuration.

                    newconfig.movedShifts = new DirectionList(dir, shifts);
                                    
                    done = newconfig.tryShifts(pieceNum, done, newconfig.movedShifts);
                }                   
            }   
        
        return done;
    }   */
    
        
    /** Is this configuration solved?  It's solved if the goal piece is in the goal
        location. */

    public boolean solved()
    {
        for (PieceList pl = pieces; pl != null; pl = pl.next)
            if (pl.piece.solved())
                return true;
                
        return false;
    }
    
    /** Print the solution.  We recursively call printSolution2 on all predecessor positions,
        until there are no more.  Then we know we're at the starting configuration,
        and can print it.  On the way back up the recursion stack, print out the
        move sequence, which we stored in the configurations.
        Successive shifts of the same piece are counted (and printed) as a single move. */

    public void printSolution()
    {
        printSolution2();

        System.out.println("\nEND SOLUTION");
    }
    
    public void printSolution2()
    {
        if (predecessor == null)
        {
            // This is our starting configuration.  Print some useful debugging & visualization stuff,
            // then print the formal solution.
            
            System.out.println("\nFound a solution!  Length is " + maxDistance + ".  Here's what it looks like:");

            prettyPrint();

            System.out.println("\n\nBEGIN SOLUTION");

            pieces.print();
            
            System.out.print("-\n" + maxDistance + "\n-\n");
        }
        else
        {
            predecessor.printSolution2();
            
            System.out.print("Move piece " + movedPiece + " ");
            movedShifts.print();
            System.out.println();
        }
    }
    
    
    /** Print a configuration for human consumption, in a nice ASCII grid layout. */
    
    public void prettyPrint()
    {
        char table[][] = new char[2 * spec.boxShape.width() + 3][spec.boxShape.height() + 2];

        for (int x = 0; x < 2 * spec.boxShape.width() + 3; ++x)
            for (int y = 0; y < spec.boxShape.height() + 2; ++y)
                table[x][y] = '*';
                
        PieceList pl;
        int pieceNum;
        
        for (int x = 0; x < spec.boxShape.width(); ++x)
            for (int y = 0; y < spec.boxShape.height(); ++y)
            {
                if (!spec.boxShape.occupies(x, y))
                {
                    table[2 * x + 1][y + 1] = table[2 * x + 3][y + 1] = ' ';
                    table[2 * x + 2][y + 1] = '.';
                }
                
                for (pl = pieces, pieceNum = 0; pl != null; pl = pl.next, ++pieceNum)
                    if (pl.piece.occupies(x, y))
                    {
                        table[2 * x + 2][y + 1] = (char) ('0' + pieceNum);
                        break;
                    }
            }
                
        for (int y = 0; y < spec.boxShape.height() + 2; ++y)
        {
            for (int x = 0; x < 2 * spec.boxShape.width() + 3; ++x)
                System.out.print(table[x][y]);

            System.out.println();               
        }
    }
    
        
    /** Can a piece of a given shape be placed into this configuration with upper
        left corner at (x, y)?
        If so, place it there & return true.  Otherwise return false. */

    public boolean placePiece(int shapeIndex, Point where)
    {
        // we can only use the goal piece once.
        
        if (shapeIndex == spec.goalIndex && usesGoal)
            return false;

        // ... and don't bother starting it in the goal position!
        
        if (shapeIndex == spec.goalIndex && where.equals(spec.goalLocation))
            return false;
            
        Shape shape = spec.pieceShapes[shapeIndex];

        for (int i = 0; i < shape.width(); ++i)
            for (int j = 0; j < shape.height(); ++j)
                if (shape.occupies(i, j) && occupied(where.x + i, where.y + j))
                    return false;

        pieces = new PieceList(new Piece(shapeIndex, where), pieces);
        
        if (shapeIndex == spec.goalIndex)
            usesGoal = true;
            
        return true;
    }
    
    
    /** Remove the last piece added to this configuration. */
    
    public void removePiece()
    {
        if (pieces.piece.isGoal())
            usesGoal = false;
            
        pieces = pieces.next;
    }
    

    /** Is the square (x, y) occupied by any piece in this configuration, or by a
        box wall?  (Anything outside the box count as a box wall too.) */

    public boolean occupied(int x, int y)
    {
        if (x < 0 || y < 0 || x >= spec.boxShape.width() || y >= spec.boxShape.height() || spec.boxShape.occupies(x, y))
            return true;
            
        for (PieceList pl = pieces; pl != null; pl = pl.next)
            if (pl.piece.occupies(x, y))
                return true;
        
        return false;
    }
    
    
    /** Does this configuration have the same piece shapes and locations as another
        configuration? */

    public boolean equals(Configuration config)
    {
        for (PieceList pl1 = pieces, pl2 = config.pieces; pl1 != null || pl2 != null; pl1 = pl1.next, pl2 = pl2.next)
        {
            if (pl1 == null || pl2 == null)
                return false;                       // one list was shorter than the other
                
            if (!pl1.piece.equals(pl2.piece))
                return false;                       // corresponding pieces don't match
        }
        
        return true;
    }
};
